"""Class definition for Copilot client"""
import logging
import re

from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from ..base_browser import BaseBrowser


class CopilotClient(BaseBrowser):
    """
    PiClient class to interact with Pi.
    It helps you to connect to https://copilot.microsoft.com/.
    Apart from core functionality Copilot supports web search.
    It is not possible to regenerate a response by using Copilot
    """

    def __init__(self, **kwargs):
        super().__init__(
            client_name="Copilot",
            url="https://copilot.microsoft.com",
            credential_check=False,
            timeout_dur=30,
            **kwargs,
        )

    def login(self, username: str = None, password: str = None):
        """
        Performs the login process with the provided username and password.
        You don't need to login to use Pi

        This function operates on the login page.
        It finds and clicks the login button,
        fills in the email and password textboxes

        Args:
            username (str): The username to be entered.
            password (str): The password to be entered.

        Returns:
            bool : True
        """
        logging.info("Login is not provided for Copilot at the moment.")
        return True

    def postload_custom_func(self) -> None:
        """Copilot requires to accept privacy terms, the cookie below provides the answer."""
        self.browser.add_cookie({"name": "BCP", "value": "AD=0&AL=0&SM=0"})
        self.browser.get(self.url)
        return

    def is_ready_to_prompt(self, text_area, shadow_element) -> bool:
        """
        Checks if the Copilot is ready to be prompted.
        The indication for an ongoing message generation process
        is a disabled send button. The indication for no input is the same
        disabled button. Therefore we put a dummy dot into the textarea
        and we are left with the only reason for the button to be disabled,
        that is, a message being generated.

        Returns:
            bool : return if the system is ready to be prompted.
        """
        text_area.send_keys(".")

        button = self.find_or_fail(
            By.CLASS_NAME, self.markers.button_cq, dom_element=shadow_element
        )
        button = self.find_or_fail(By.TAG_NAME, "button", dom_element=button)
        if not button:
            return False

        self.wait_object.until(EC.element_to_be_clickable(button))

        # Then, we clear the text area to make space for new interacton :)
        text_area.send_keys(Keys.CONTROL + "a", Keys.DELETE)
        return True

    def interact(self, prompt: str):
        """Sends a prompt and retrieves the answer from the ChatGPT system.

        This function interacts with the PI.
        It takes the prompt as input and sends it to the system.
        The prompt may contain multiple lines separated by '\\n'.
        In this case, the function simulates pressing SHIFT+ENTER for each line.
        Upon arrival of the interaction, the function waits for the answer.
        Once the response is ready, the function will return the response.

        Args:
            prompt (str): The interaction text.

        Returns:
            Dict[str]: The generated answer and references.
        """
        main_area = self.find_or_fail(
            By.TAG_NAME, self.markers.main_area_tq, return_shadow=True
        )
        action_bar = self.find_or_fail(
            By.ID, self.markers.action_bar_iq, dom_element=main_area, return_shadow=True
        )
        input_bar = self.find_or_fail(
            By.CLASS_NAME, self.markers.input_bar_cq, dom_element=action_bar
        ).children()[0].shadow_root
        text_area = self.find_or_fail(
            By.ID, self.markers.textarea_iq, dom_element=input_bar
        )

        if not text_area:
            logging.error("Unable to locate text area, interaction fails.")
            return ""

        for each_line in prompt.split("\n"):
            text_area.send_keys(each_line)
            text_area.send_keys(Keys.SHIFT + Keys.ENTER)

        # Click enter and send the message
        text_area.send_keys(Keys.ENTER)

        if not self.is_ready_to_prompt(text_area, action_bar):
            logging.info("Cannot retrieve the answer, something is wrong")
            return ""

        # Shadow roots, aren't they amazing!

        resp = self.find_or_fail(
            By.ID, self.markers.con_main_iq, dom_element=main_area, return_shadow=True
        )
        resp = self.find_or_fail(
            By.CSS_SELECTOR,
            self.markers.con_chat_sq,
            dom_element=resp,
            return_type="last",
            return_shadow=True,
        )
        resp = self.find_or_fail(
            By.CLASS_NAME,
            self.markers.con_resp_cq,
            dom_element=resp,
            return_shadow=True,
        )
        resp = self.find_or_fail(
            By.CSS_SELECTOR,
            self.markers.con_msg_sq,
            dom_element=resp,
            return_type="last",
            return_shadow=True,
        )
        resp = self.find_or_fail(
            By.CSS_SELECTOR, self.markers.con_ins_sq, dom_element=resp
        )
        resp = self.find_or_fail(
            By.CLASS_NAME, self.markers.con_last_cq, dom_element=resp
        )
        # TO DO - return text and ref as dictionary instead of just returning text
        # text_elems = list(
        #     filter(lambda x: re.search(r"^\d\n", x.text) is None, resp.children())
        # )

        # refs = []
        # if len(text_elems) < len(resp.children()):
        #     refs = resp.children()[-1]
        #     refs = [
        #         i.get_attribute("href")
        #         for i in refs.find_elements(By.CLASS_NAME, self.markers.ref_cq)
        #     ]

        text = "".join([elem.text for elem in resp.children()])
        # Fix citations
        text = re.sub(r"\n(\d{1,2})", r"[\g<1>]", text)

        if not text:
            return ""
        logging.info("Answer is ready")
        self.log_chat(prompt=prompt, answer=text)
        return text

    def reset_thread(self) -> bool:
        """
        Function to close the current thread and start new one

        Returns:
            bool: False always, it is not possible to reset in Pi.
        """
        main_area = self.find_or_fail(
            By.TAG_NAME, self.markers.main_area_tq, return_shadow=True
        )
        if not main_area:
            return False

        action_bar = self.find_or_fail(
            By.ID, self.markers.action_bar_iq, dom_element=main_area, return_shadow=True
        )
        if not action_bar:
            return False

        new_chat_button = self.find_or_fail(
            By.CLASS_NAME, self.markers.new_chat_cq, dom_element=action_bar
        )
        if not new_chat_button:
            return False
        new_chat_button.click()

        return True

    def switch_model(self, model_name: str):
        """
        Switch the model. (style)

        Args:
            model_name: str = The name of the model

        Returns:
            bool: True on success, False on fail
        """
        main_area = self.find_or_fail(
            By.TAG_NAME, self.markers.main_area_tq, return_shadow=True
        )

        con_area = self.find_or_fail(
            By.ID, self.markers.con_main_iq, dom_element=main_area, return_shadow=True
        )

        wel_area = self.find_or_fail(
            By.CSS_SELECTOR,
            self.markers.welcome_tq,
            dom_element=con_area,
            return_shadow=True,
        )

        tone_area = self.find_or_fail(
            By.CSS_SELECTOR,
            self.markers.tone_tq,
            dom_element=wel_area,
            return_shadow=True,
        )

        models = dict(
            map(
                lambda x: (x.text.split("\n")[-1], x),
                self.find_or_fail(
                    By.CSS_SELECTOR, "button", dom_element=tone_area, return_type="all"
                ),
            )
        )

        button = models.get(model_name, None)
        if button is None:
            logging.error("Model %s has not found", model_name)
            logging.error("Available models are: %s", str(models.keys()))
            return False
        button.click()

        verification = button.get_attribute("aria-checked")
        if verification != "true":
            logging.error("Model switch to %s is unsuccessful", model_name)
            return False

        logging.info("Switched to %s", model_name)
        return True

    def toggle_plugin(self, plugin_name: str):
        """Toggles the status of the plugin.
        In order to use plugins, search plugin should be active.

        Args:
            plugin_name (str): The name of the plugin

        Returns:
            _type_: True if toggle operation is successful, False if operation fails.
        """

        main_area = self.find_or_fail(
            By.TAG_NAME, self.markers.main_area_tq, return_shadow=True
        )

        side_panel = self.find_or_fail(
            By.CSS_SELECTOR,
            self.markers.side_tq,
            dom_element=main_area,
            return_shadow=True,
        )

        self.find_or_fail(
            By.CSS_SELECTOR, "button", dom_element=side_panel, return_type="last"
        ).click()

        plugin_panel = self.find_or_fail(
            By.CSS_SELECTOR,
            self.markers.plugin_tq,
            dom_element=side_panel,
            return_shadow=True,
        )

        plugins = self.find_or_fail(
            By.CLASS_NAME,
            self.markers.p_control_cq,
            dom_element=plugin_panel,
            return_type="all",
        )

        inputs = self.find_or_fail(
            By.CSS_SELECTOR, "input", dom_element=plugin_panel, return_type="all"
        )

        plugin_map = dict(map(lambda x: (x[0].text, x[1]), zip(plugins, inputs)))

        if plugin_name not in plugin_map:
            logging.error("Plugin %s has not found", plugin_name)
            logging.error("Available plugins are: %s", str(plugin_map.keys()))
            return

        if not plugin_map["Search"].is_selected():
            logging.info("Search is disabled, activating it")
            plugin_map["Search"].click()
            logging.info("Search is enabled")

        p_element = plugin_map[plugin_name]

        logging.info(
            "The plugin is currently %s",
            ["disabled", "enabled"][p_element.is_selected()],
        )
        p_element.click()
        logging.info(
            "The plugin is now %s", ["disabled", "enabled"][p_element.is_selected()]
        )
        return True

    def regenerate_response(self):
        raise NotImplementedError("Copilot doesn't provide response regeneration")



"""Class definition for HuggingChat client"""
import logging

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from ..base_browser import BaseBrowser


class HuggingChatClient(BaseBrowser):
    """
    HuggingChatClient class to interact with HuggingChat.
    It helps you to conncet to https://huggingface.co/chat/ and login.
    Apart from core functionality HuggingChat supports web search.
    It is not possible to regenerate a response by using HuggingChat
    """

    def __init__(self, **kwargs):
        super().__init__(
            client_name="HuggingChat",
            url="https://huggingface.co/chat/",
            timeout_dur=45,
            **kwargs,
        )

    def login(self, username: str, password: str):
        """
        Performs the login process with the provided username and password.

        This function operates on the login page.
        It finds and clicks the login button,
        fills in the email and password textboxes

        Args:
            username (str): The username to be entered.
            password (str): The password to be entered.

        Returns:
            bool : True if login is successful
        """

        # Find login button, click it
        login_button = self.wait_until_appear(By.XPATH, self.markers.login_xq)
        login_button.submit()
        logging.info("Clicked login button")

        # Find email textbox, enter e-mail
        email_box = self.wait_until_appear(By.XPATH, self.markers.username_xq)
        email_box.send_keys(username)
        logging.info("Filled username/email")

        # Find password textbox, enter password
        pass_box = self.wait_until_appear(By.XPATH, self.markers.password_xq)
        pass_box.send_keys(password)
        logging.info("Filled password box")

        # Click continue
        a_login_button = self.wait_until_appear(By.XPATH, self.markers.a_login_xq)
        a_login_button.click()
        logging.info("Clicked login button")
        return True

    def interact(self, prompt: str):
        """Sends a prompt and retrieves the answer from the ChatGPT system.

        This function interacts with the HuggingChat.
        It takes the prompt as input and sends it to the system.
        The prompt may contain multiple lines seperated by '\\n'.
        In this case, the function simulates pressing SHIFT+ENTER for each line.
        Upon arrival of the interaction, the function waits for the answer.
        Once the response is ready, the function will return the response.

        Args:
            prompt (str): The interaction text.

        Returns:
            str: The generated answer.
        """

        text_area = self.find_or_fail(By.XPATH, self.markers.textarea_xq)
        if not text_area:
            return ""

        for each_line in prompt.split("\n"):
            text_area.send_keys(each_line)
            text_area.send_keys(Keys.SHIFT + Keys.ENTER)
        text_area.send_keys(Keys.RETURN)
        logging.info("Message sent, waiting for response")
        self.wait_until_disappear(By.XPATH, self.markers.stop_gen_xq)
        answer = self.find_or_fail(
            By.XPATH, self.markers.chatbox_xq, return_type="last"
        )
        if not answer:
            return ""
        logging.info("Answer is ready")
        self.log_chat(prompt=prompt, answer=answer.text)
        return answer.text

    def reset_thread(self):
        """Function to close the current thread and start new one"""
        self.browser.get(self.url)
        return True

    def toggle_search_web(self):
        """Function to enable/disable web search feature"""
        search_web_toggle = self.find_or_fail(By.XPATH, self.markers.search_xq)
        if not search_web_toggle:
            return
        search_web_toggle.click()
        status = search_web_toggle.get_attribute("aria-checked")
        status = status == "true"
        logging.info("Search web is %s", ["disabled", "enabled"][status])
        return status

    def switch_model(self, model_name: str):
        """
        Switch the model.

        Args:
            model_name: str = The name of the model

        Returns:
            bool: True on success, False on fail
        """
        model_button = self.find_or_fail(By.XPATH, self.markers.model_xq)
        if not model_button:
            return False
        model_button.click()

        self.wait_object.until(
            EC.presence_of_element_located((By.XPATH, self.markers.settings_xq))
        )
        models = self.find_or_fail(
            By.XPATH, self.markers.model_li_xq, return_type="all"
        )
        if not models:
            return False
        models = {m.text.strip(): m for m in models}

        model = models.get(model_name, None)
        if model is None:
            logging.error("Model %s has not found", model_name)
            logging.error("Available models are: %s", str(models.keys()))
            return False

        model.click()
        logging.info("Switched to %s", model_name)

        apply_button = self.find_or_fail(By.XPATH, self.markers.model_a_xq)
        if not apply_button:
            return False

        apply_button.click()
        return True

    def regenerate_response(self):
        raise NotImplementedError("HuggingChat doesn't provide response regeneration")



"""HuggingChat test"""

import time
import pytest
from talkingheads import PiClient

def test_start():
    pytest.chathead = PiClient(headless=True, verbose=True)
    assert pytest.chathead.ready, "The Client is not ready"


def test_interaction():
    time.sleep(1)
    answer = pytest.chathead.interact(
        "Without any explanation or extra information, just repeat the following: book."
    )
    assert (
        "book" in answer.lower()
    ), f'Answer is not "book.", instead it returned {answer}'


# def test_model_selection():
#     assert pytest.chathead.switch_model("SupportPi"), "Model switch failed."


def pytest_sessionfinish(session, exitstatus):
    del pytest.chathead



"""ChatGPT test"""

import pytest
from selenium.webdriver.common.by import By
from talkingheads import ChatGPTClient


def test_start():
    pytest.chathead = ChatGPTClient(headless=True, verbose=True)
    assert pytest.chathead.ready, "The Client is not ready"


def test_interaction():
    answer = pytest.chathead.interact(
        "Without any explanation or extra information, just repeat the following: book."
    )
    assert (
        "book" in answer.lower()
    ), f'Answer is not "book.", instead it returned {answer}'


def test_reset():
    assert pytest.chathead.reset_thread()
    assert (
        len(
            pytest.chathead.browser.find_elements(
                By.XPATH, pytest.chathead.markers.chatbox_xq
            )
        )
        == 0
    ), "Chat is not empty"


def test_regenerate():
    first_answer = pytest.chathead.interact(
        "Without any explanation or extra information, type three animal names."
    ).lower()
    second_answer = pytest.chathead.regenerate_response()
    assert first_answer != second_answer, "The regenerated answer is the same."


def test_custom_interactions():
    mod_text = "Lorem ipsum dolor sit amet"
    info_text = "consectetur adipiscing elit"
    pytest.chathead.set_custom_instruction("modulation", mod_text)
    applied_mod_text = pytest.chathead.get_custom_instruction("modulation")
    assert (
        applied_mod_text == mod_text
    ), f"Modulation text is different: {applied_mod_text}"
    pytest.chathead.set_custom_instruction("extra_information", info_text)
    applied_extra_text = pytest.chathead.get_custom_instruction("extra_information")
    assert (
        applied_extra_text == info_text
    ), f"Info text is different {applied_extra_text}"


def pytest_sessionfinish(session, exitstatus):
    del pytest.chathead



"""Class definition for ChatGPTClient"""

import logging
import time
from datetime import datetime

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
import selenium.common.exceptions as Exceptions

from talkingheads import BaseBrowser


class ChatGPTClient(BaseBrowser):
    """ChatGPTClient class to interact with ChatGPT"""

    def __init__(self, **kwargs):
        super().__init__(client_name="ChatGPT", url="https://chat.openai.com", **kwargs)

    def postload_custom_func(self):
        today_str = datetime.today().strftime("%Y-%m-%d")
        self.browser.execute_script(
            f"window.localStorage.setItem('oai/apps/hasSeenOnboarding/chat', {today_str});"
            "window.localStorage.setItem('oai/apps/hasUserContextFirstTime/2023-06-29', true);"
            "window.localStorage.setItem('oai/apps/announcement/customInstructions', 1694012515508);"
        )

    def pass_verification(self, max_trial: int = 10, wait_time: int = 1) -> bool:
        """
        Performs the verification process on the page if challenge is present.

        This function checks if the login page is displayed in the browser.
        In that case, it looks for the verification button.
        This process is repeated until the login page is no longer displayed.

        Returns: None
        """
        for _ in range(max_trial):
            if not self.check_login_page():
                break
            verify_button = self.browser.find_elements(By.ID, "challenge-stage")
            if len(verify_button):
                try:
                    verify_button[0].click()
                    logging.info("Clicked verification button")
                except Exceptions.ElementNotInteractableException:
                    logging.info("Verification button is not present or clickable")
            time.sleep(wait_time)
        else:
            logging.error("It is not possible to pass verification")
            return False

        return True

    def login(self, username: str, password: str):
        """
        Performs the login process with the provided username and password.

        This function operates on the login page.
        It finds and clicks the login button,
        fills in the email and password textboxes

        Args:
            username (str): The username to be entered.
            password (str): The password to be entered.

        Returns:
            None
        """

        # Find login button, click it
        login_button = self.wait_until_appear(By.XPATH, self.markers.login_xq)
        login_button.click()
        logging.info("Clicked login button")
        time.sleep(1)

        # Find email textbox, enter e-mail
        email_box = self.wait_until_appear(By.ID, "username")
        email_box.send_keys(username)
        logging.info("Filled email box")

        # Click continue
        continue_button = self.wait_until_appear(By.XPATH, self.markers.continue_xq)
        continue_button.click()
        time.sleep(1)
        logging.info("Clicked continue button")

        # Find password textbox, enter password
        pass_box = self.wait_until_appear(By.ID, "password")
        pass_box.send_keys(password)
        logging.info("Filled password box")
        # Click continue
        pass_box.send_keys(Keys.ENTER)
        time.sleep(1)
        logging.info("Logged in")

        try:
            # Pass introduction
            WebDriverWait(self.browser, 5).until(
                EC.presence_of_element_located((By.XPATH, self.markers.tutorial_xq))
            ).click()

            logging.info("Info screen passed")
        except Exceptions.TimeoutException:
            logging.info("Info screen skipped")
        except Exception as err:
            logging.error("Something unexpected: %s", err)

    def interact(self, prompt: str) -> str:
        """Sends a prompt and retrieves the answer from the ChatGPT system.

        This function interacts with the ChatGPT.
        It takes the prompt as input and sends it to the system.
        The prompt may contain multiple lines separated by '\\n'.
        In this case, the function simulates pressing SHIFT+ENTER for each line.
        Upon arrival of the interaction, the function waits for the answer.
        Once the response is ready, the function will return the response.

        Args:
            prompt (str): The interaction text.

        Returns:
            str: The generated answer.
        """

        text_area = self.browser.find_elements(By.TAG_NAME, self.markers.textarea_tq)
        if not text_area:
            logging.info("Unable to locate text area tag. Switching to ID search")
            text_area = self.browser.find_elements(By.ID, self.markers.textarea_iq)
        if not text_area:
            raise RuntimeError(
                "Unable to find the text prompt area. Please raise an issue with verbose=True"
            )

        text_area = text_area[0]

        for each_line in prompt.split("\n"):
            text_area.send_keys(each_line)
            text_area.send_keys(Keys.SHIFT + Keys.ENTER)
        text_area.send_keys(Keys.RETURN)
        logging.info("Message sent, waiting for response")
        self.wait_until_disappear(By.XPATH, self.markers.wait_xq)
        answer = self.find_or_fail(
            By.XPATH, self.markers.chatbox_xq, return_type="last"
        )
        if not answer:
            logging.error("There is no answer, something is wrong")
            return ""

        logging.info("Answer is ready")
        self.log_chat(prompt=prompt, answer=answer.text)

        return answer.text

    def reset_thread(self) -> bool:
        """Function to close the current thread and start new one"""
        self.browser.get(self.url)
        return True

    def regenerate_response(self) -> str:
        """
        Clicks the response to generate a new one and returns the new answer.

        Args:
            None

        Returns:
            str
        """
        regen_button = self.find_or_fail(
            By.XPATH, self.markers.regen_xq, return_type="all"
        )
        if not regen_button:
            return ""
        regen_button[2].click()
        logging.info("Clicked regenerate button")

        self.wait_until_disappear(By.XPATH, self.markers.wait_xq)
        answer = self.find_or_fail(
            By.XPATH, self.markers.chatbox_xq, return_type="last"
        )
        if not answer:
            return ""

        logging.info("New answer is ready")
        self.log_chat(answer=answer.text, regenerated=True)
        return answer.text

    def switch_model(self, model_name: str) -> bool:
        """
        Switch the model for ChatGPT+ users.

        Args:
            model_name: str = The name of the model, either GPT-3.5 or GPT-4

        Returns:
            bool: True on success, False on fail
        """
        if model_name in ["GPT-3.5", "GPT-4"]:
            logging.info("Switching model to %s", model_name)
            try:
                self.browser.find_element(
                    By.XPATH, self.markers.gpt_xq.format(model_name)
                ).click()
                return True
            except Exceptions.NoSuchElementException:
                logging.error("Button is not present")
        else:
            logging.error("Model name is not ")
        return False

    def open_custom_instruction_tab(self) -> bool:
        """Opens the modal to access custom interactions.

        Returns:
            bool: True if the process is successful, False otherwise
        """

        menu_button = self.find_or_fail(By.XPATH, self.markers.menu_xq)
        menu_button.click()
        custom_button = self.find_or_fail(By.XPATH, self.markers.custom_xq)
        custom_button.click()
        custom_tutorial = self.find_or_fail(
            By.XPATH, self.markers.custom_tutorial_xq, fail_ok=True
        )
        if custom_tutorial:
            custom_tutorial.click()

        custom_switch = self.find_or_fail(By.XPATH, self.markers.custom_toggle_xq)
        if not custom_switch:
            return False

        # If disabled, enable custom interactions
        if custom_switch.get_attribute("data-state") == "checked":
            logging.info("Custom instructions is enabled")
        else:
            custom_switch.click()
        time.sleep(0.2)

        return True

    def get_custom_instruction(self, mode: str) -> str:
        """Gets custom instructions

        Args:
            mode (str): Either 'extra_information' or 'modulation'. Check OpenAI help pages.
        """
        if mode not in ["extra_information", "modulation"]:
            logging.error(
                "Given mode is unrecognized. Either provide extra_information or modulation"
            )
            return ""
        if not self.open_custom_instruction_tab():
            return ""
        text_areas = self.find_or_fail(
            By.XPATH, self.markers.custom_textarea_xq, return_type="all"
        )
        text = text_areas[{"extra_information": 0, "modulation": 1}[mode]].text
        logging.info("Custom instruction is obtained: %s", text)

        save_button = self.find_or_fail(By.XPATH, self.markers.custom_cancel_xq)
        save_button.click()
        return text

    def set_custom_instruction(self, mode: str, instruction: str):
        """Sets custom instructions

        Args:
            mode (str): Either 'extra_information' or 'modulation'. Check OpenAI help pages.
            instruction (str): _description_
        """
        if not self.open_custom_instruction_tab():
            return False

        WebDriverWait(self.browser, 5).until(
            EC.presence_of_element_located((By.XPATH, self.markers.custom_textarea_xq))
        )
        text_areas = self.find_or_fail(
            By.XPATH, self.markers.custom_textarea_xq, return_type="all"
        )
        text_area = text_areas[{"extra_information": 0, "modulation": 1}[mode]]

        text_area.send_keys(Keys.CONTROL + "a")
        time.sleep(0.1)
        text_area.send_keys(Keys.DELETE)
        time.sleep(0.1)
        text_area.send_keys(instruction)
        time.sleep(0.1)
        text_area.send_keys(" ")
        logging.info("Custom instruction-%s has provided", mode)

        save_button = self.find_or_fail(By.XPATH, self.markers.custom_save_xq)
        save_button.click()
        return True



"""Multiagent"""

import asyncio
import logging

import yaml
from talkingheads import model_library

client_map = {
    "ChatGPT": model_library.ChatGPTClient,
    "HuggingChat": model_library.HuggingChatClient,
    "Gemini": model_library.GeminiClient,
    "Pi": model_library.PiClient,
}


class TalkingHeads:
    """An interface for talking heads"""

    def __init__(self, configuration):
        with open(configuration) as fd:
            self.config = yaml.safe_load(fd)

        self.agent_list = []
        self.agent_map = {}
        self.master_head = None

        for conf in enumerate(self.config.nodes):
            client = self.open_client(conf)
            self.agent_map[conf.tag] = client
            self.agent_list.append(client)

    def open_client(self, config):
        return 1

    def set_master_head(self, head_number: int = None, head_name: str = None) -> bool:
        if not head_name or head_number:
            logging.error("Either provide head_number or head_name")
            return False

        head = self.agent_map.get(head_name)
        if not head:
            logging.error("No such head %s exists.", head_name)
            return False
        self.master_head = head
        return True

    def interact(
        self, question: str, head_number: int = None, head_name: str = None
    ) -> str:
        """interact with the given head"""
        if head_number:
            client = self.agent_list[head_number]
        elif head_name:
            client = self.agent_map[head_name]
        else:
            raise NameError("Either head number or tag should be provided.")

        response = client.interact(question)
        return response

    async def broadcast(self, prompt, exclude_master=False):
        """Sends message to all and returns back."""
        heads = list(self.agent_map.values())
        if exclude_master:
            heads.remove(self.master_head)

        responses = await asyncio.gather(
            *[asyncio.create_task(agent.interact(prompt)) for agent in heads]
        )
        return responses

    async def reset_all_threads(self):
        """reset heads for the given number"""
        reset_status = await asyncio.gather(
            *[asyncio.create_task(agent.reset_thread()) for agent in self.agent_map]
        )
        return all(reset_status)

    async def aggregate(self, prompt: str, agg_objective: str):
        """This function collects a number of responses and aggregates them.
        The aggregation can differ in logical means. Some examples:
        - Compose a new message
        - Select the best message
        - Select the worst message
        - Find the most reoccuring number
        - Check if all the answers are the semantically similar

        Not all chat models can read lengthy inputs, remind your chat models to keep it short!

        Args:
            prompt (str): The prompt to pass on model swarm.
            agg_objective (str): The prompt to pass on master head to apply the aggregation.

        Returns:
            (str): The response of the master head.
        """
        if self.master_head is None:
            raise RuntimeError("Set master head before voting operation")

        answers = await self.broadcast(prompt, exclude_master=True)
        answers = " \n".join(
            [f"{idx})\n{answer}" for idx, answer in enumerate(answers)]
        )
        final_prompt = f"""
        {agg_objective}
        Here are the options:
        {answers}
        """
        final_answer = self.master_head(final_prompt)
        return final_answer



"""Class definition for Claude client"""
import logging

from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from ..base_browser import BaseBrowser


class ClaudeClient(BaseBrowser):
    """
    ClaudeClient class to interact with Claude.
    It helps you to conncet to https://claude.ai/.
    Apart from core functionality Claude supports web search.
    It is not possible to regenerate a response by using Claude
    """

    def __init__(self, **kwargs):
        super().__init__(
            client_name="Claude",
            url="https://claude.ai",
            credential_check=False,
            timeout_dur=30,
            **kwargs,
        )

    def login(self, _1: str = None, _2: str = None):
        """
        It is only possible to login Claude manually.

        Args:
            _1 (str): Unused parameter 1.
            _2 (str): Unused parameter 2.

        Returns:
            bool : True
        """
        logging.info("Login is not provided for Claude at the moment.")
        return True

    def postload_custom_func(self) -> None:
        """In order to continue chat screen, a click on start button is required"""
        start_button = self.wait_until_appear(By.XPATH, self.markers.start_button_xq)
        if not start_button:
            logging.error("Post-load has failed.")
            return False
        start_button.click()
        return

    def is_ready_to_prompt(self) -> bool:
        """
        Checks if the Claude is ready to be prompted.
        The indication for an ongoing message generation process
        is a disabled send button. The indication for no input is the same
        disabled button. Therefore we put a dummy dot into the textarea
        and we are left with the only reason for the button to be disabled,
        that is, a message being generated.

        Returns:
            bool : return if the system is ready to be prompted.
        """
        text_area = self.find_or_fail(By.CLASS_NAME, self.markers.textarea_cq)
        text_area.send_keys(".")

        button = self.find_or_fail(By.XPATH, self.markers.send_button_xq)
        if not button:
            return False

        self.wait_object.until(EC.element_to_be_clickable(button))

        # Then, we clear the text area to make space for new interacton :)
        text_area.send_keys(Keys.CONTROL + "a", Keys.DELETE)
        return True

    def interact(self, prompt: str):
        """Sends a prompt and retrieves the answer from the ChatGPT system.

        This function interacts with the Claude.
        It takes the prompt as input and sends it to the system.
        The prompt may contain multiple lines separated by '\\n'.
        In this case, the function simulates pressing SHIFT+ENTER for each line.
        Upon arrival of the interaction, the function waits for the answer.
        Once the response is ready, the function will return the response.

        Args:
            prompt (str): The interaction text.

        Returns:
            Dict[str]: The generated answer and references.
        """

        text_area = self.find_or_fail(By.CLASS_NAME, self.markers.textarea_cq)

        if not text_area:
            logging.error("Unable to locate text area, interaction fails.")
            return ""

        for each_line in prompt.split("\n"):
            text_area.send_keys(each_line)
            text_area.send_keys(Keys.SHIFT + Keys.ENTER)

        # Click enter and send the message
        text_area.send_keys(Keys.ENTER)

        if not self.is_ready_to_prompt():
            logging.info("Cannot retrieve the answer, something is wrong")
            return ""

        answer = self.find_or_fail(
            By.XPATH,
            self.markers.chatarea_xq,
            return_type="last",
        )
        if not answer:
            return ""
        logging.info("Answer is ready")
        self.log_chat(prompt=prompt, answer=answer.text)
        return answer.text

    def reset_thread(self) -> bool:
        """
        Function to close the current thread and start new one

        Returns:
            bool: True if reset is successful, false otherwise.
        """
        text_area = self.find_or_fail(By.CLASS_NAME, self.markers.textarea_cq)
        text_area.send_keys(Keys.CONTROL + "K")
        start_button = self.find_or_fail(By.XPATH, self.markers.start_button_xq)
        if not start_button:
            return False
        start_button.click()

        return True

    def switch_model(self, _: str):
        raise NotImplementedError("Claude only has one model")

    def regenerate_response(self):
        regen_button = self.find_or_fail(By.XPATH, self.markers.regen_xq)
        if not regen_button:
            return ""
        regen_button.click()

        if not self.is_ready_to_prompt():
            logging.info("Cannot retrieve the answer, something is wrong")
            return ""

        answer = self.find_or_fail(
            By.XPATH,
            self.markers.chatarea_xq,
            return_type="last",
        )
        if not answer:
            return ""

        logging.info("Answer is ready")
        self.log_chat(answer=answer.text, regenerated=True)
        return answer.text



"""HuggingChat test"""

import time
import pytest
from selenium.webdriver.common.by import By
from talkingheads import HuggingChatClient


def test_start():
    pytest.chathead = HuggingChatClient(headless=True, verbose=True)
    assert pytest.chathead.ready, "The Client is not ready"


def test_model_selection():
    assert pytest.chathead.switch_model(
        "meta-llama/Llama-2-70b-chat-hf"
    ), "Model switch failed."


def test_interaction():
    time.sleep(1)
    answer = pytest.chathead.interact(
        "Without any explanation or extra information, just repeat the following: book."
    )
    assert (
        "book" in answer.lower()
    ), f'Answer is not "book.", instead it returned {answer}'


def test_reset():
    assert pytest.chathead.reset_thread()
    assert (
        len(
            pytest.chathead.browser.find_elements(
                By.XPATH, pytest.chathead.markers.chatbox_xq
            )
        )
        == 0
    ), "Chat is not empty"


def pytest_sessionfinish(_session, _exitstatus):
    """exits"""
    del pytest.chathead



"""Model inits"""

from .chatgpt import ChatGPTClient
from .huggingchat import HuggingChatClient
from .gemini import GeminiClient
from .pi import PiClient
from .copilot import CopilotClient
from .claude import ClaudeClient



'''Multiagent'''

import yaml
import asyncio
import model_library

client_map = {
    'ChatGPT' : model_library.ChatGPTClient,
    'HuggingChat' : model_library.HuggingChatClient,
    'Gemini' : model_library.GeminiClient,
    'Pi' : model_library.PiClient
}

class TalkingHeads:
    """An interface for talking heads"""
    def __init__(self, configuration):
        with open(configuration) as fd:
            self.config = yaml.safe_load(fd)

        self.agent_list = []
        self.agent_map = {}

        for conf in enumerate(self.config.nodes):
            client = self.open_client(conf)
            self.agent_map[conf.tag] = client
            self.agent_list.append(client)

    def open_client(self, config):
        return 1

    def interact(self, question : str, head_number : int = None, head_name : str = None) -> str:
        """interact with the given head"""
        if head_number:
            client = self.agent_list[head_number]
        elif head_name:
            client = self.agent_map[head_name]
        else:
            raise NameError('Either head number or tag should be provided.')

        response = client.interact(question)
        return response

    async def broadcast(self, prompt):
        '''Sends message to all and returns back.'''
        responses = await asyncio.gather(
            *[
                asyncio.create_task(agent.interact(prompt))
                for agent in self.agent_map
            ]
        )
        return responses

    async def reset_all_threads(self):
        """reset heads for the given number"""
        reset_status = await asyncio.gather(
            *[
                asyncio.create_task(agent.reset_thread())
                for agent in self.agent_map
            ]
        )
        return all(reset_status)

    async def voting(self, prompt, voting_objective):
        """votes"""
        answers = await self.broadcast(prompt)
        voting_prompt = ' \n'.join([
            f'{idx})\n{answer}'
            for idx, answer in answers
        ])
        voting_prompt = f'''
        {voting_objective}
        Here are the options:
        {voting_prompt}
        Please provide your reasoning and at the end of your answer, provide your preference with the number indicator.
        '''
        selection = await self.broadcast(voting_prompt)
        best_option = [re.search('\d', i.split('\n')[-1]) for i in selection]
        best_option = 
        return all(reset_status)




"""Utility functions of talkingheads library"""
import re
import subprocess
import logging
from typing import Union

from undetected_chromedriver import find_chrome_executable


def detect_chrome_version(version_num: int = None) -> Union[int, None]:
    """
    Detects the Google Chrome version on Linux and macOS machines.

    Parameters:
        version_num (int, optional): The chromedriver version number. Default: None.

    Returns:
        int: The detected Google Chrome version number.

    Note:
    - If version_num is provided, it will be returned without any detection.
    - Uses subprocess to execute the 'google-chrome --version' command for detection.
    - If the command output doesn't match the expected format, it returns None.
    - Logs information about the detected or default version using the logging module.
    """

    if version_num:
        logging.debug("Version number is provided: %d", version_num)
        return version_num

    chrome_path = find_chrome_executable()

    out = subprocess.check_output([chrome_path, "--version"])
    out = re.search(r"Google\s+Chrome\s+(\d{3})", out.decode())

    if not out:
        logging.error("There was an error obtaining Chrome version")
        return None

    version_num = int(out.group(1))
    logging.info("The version is %d", version_num)
    return version_num


save_func_map = {
    "csv": "to_csv",
    "h5": "to_hdf",
    "html": "to_html",
    "json": "to_json",
    "orc": "to_orc",
    "pkl": "to_pkl",
    "xlsx": "to_xlsx",
    "xml": "to_xml",
}



"""Class definition for ChatGPT_Client"""

import abc
import os
import logging
from datetime import datetime
from typing import Union

import undetected_chromedriver as uc
import pandas as pd

from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.remote.webelement import WebElement
import selenium.common.exceptions as Exceptions

from .object_map import markers
from .utils import detect_chrome_version, save_func_map


class BaseBrowser:
    """
    BaseBrowser class to provide utility function and flow for LLMs

    Args:
        client_name (str): A string representing the name of the client.
        url (str): The URL to be used as an entrypoint.
        uname_var (str): The username environment variable, to enable multiple agents.
        pwd_var (str): The password environment variable, to enable multiple agents.
        username (str, optional): The username to auth. Deprecated. Use environment variables instead.
        password (str, optional): The password to auth. Deprecated. Use environment variables instead.
        headless (bool, optional): A boolean to enable/disable headless mode in driver. Default: True.
        cold_start (bool, optional): If set, it will return after opening the browser. Default: False.
        incognito (bool, optional): A boolean to set incognito mode. Default: True.
        driver_arguments (list, optional): A list of additional arguments to be passed to the driver. Default: None.
        driver_version (int, optional): The version of the chromedriver. Default: None.
        auto_save (bool, optional): A boolean to enable/disable automatic saving. Default: False.
        save_path (str, optional): The file path to save chat logs. Default: None.
        verbose (bool, optional): A boolean to enable/disable logging. Default: False.
        credential_check (bool, optional): A boolean to enable/disable credential check. Default: True.
        skip_login (bool, optional): If True, skips the login procedure. Default: False.
        user_data_dir (str, optional): The directory path to user profile. Default: None.
        uc_params (dict, optional): Parameters for uc.Chrome().
            Some examples : driver_executable_path, browser_executable_path

    Returns:
        BaseBrowser: The base driver object.
    """

    def __init__(
        self,
        client_name: str,
        url: str,
        uname_var: Union[str, None] = None,
        pwd_var: Union[str, None] = None,
        username: Union[str, None] = None,
        password: Union[str, None] = None,
        headless: bool = True,
        cold_start: bool = False,
        incognito: bool = True,
        driver_arguments: list = None,
        driver_version: int = None,
        timeout_dur: int = 15,
        auto_save: bool = False,
        save_path: str = None,
        verbose: bool = False,
        credential_check: bool = True,
        skip_login: bool = False,
        user_data_dir: str = None,
        uc_params: dict = None,
    ):
        self.client_name = client_name
        self.markers = markers[client_name]
        self.url = url
        self.uname_var = uname_var or f"{client_name}_UNAME"
        self.pwd_var = pwd_var or f"{client_name}_PWD"
        self.headless = headless
        self.ready = False
        self.auto_save = auto_save

        if credential_check:
            if username or password:
                logging.warning(
                    "The username and password parameters are deprecated and will be removed soon."
                    " Please adjust your environment variables to pass username and password."
                )

            username = username or os.environ.get(self.uname_var)
            password = password or os.environ.get(self.pwd_var)

            if not username:
                raise NameError(
                    f"Either provide username or set the environment variable {self.uname_var}"
                )

            if not password:
                raise NameError(
                    f"Either provide password or set the environment variable {self.pwd_var}"
                )

        if verbose:
            logging.getLogger().setLevel(logging.INFO)
            logging.info("Verbose mode active")
        options = uc.ChromeOptions()
        options.headless = self.headless
        if incognito:
            options.add_argument("--incognito")
        if driver_arguments:
            for _arg in driver_arguments:
                options.add_argument(_arg)

        logging.info("Loading undetected Chrome")
        uc_params = uc_params or {}
        self.browser = uc.Chrome(
            user_data_dir=user_data_dir,
            options=options,
            headless=headless,
            version_main=detect_chrome_version(driver_version),
            **uc_params,
        )
        self.wait_object = WebDriverWait(self.browser, timeout_dur)
        agent = self.browser.execute_script("return navigator.userAgent")
        self.browser.execute_cdp_cmd(
            "Network.setUserAgentOverride", {"userAgent": agent.replace("Headless", "")}
        )
        self.browser.set_page_load_timeout(15)

        logging.info("Loaded undetected Chrome")
        logging.info("Opening %s", self.client_name)

        self.preload_custom_func()
        self.browser.get(self.url)
        if cold_start:
            return

        self.postload_custom_func()
        if not self.pass_verification():
            raise RuntimeError("Verification failed, please check your connection.")

        if not skip_login:
            self.login(username, password)

        logging.info("%s is ready to interact", self.client_name)
        self.ready = True
        self.chat_history = pd.DataFrame(columns=["role", "is_regen", "content"])
        self.set_save_path(save_path)

    def __del__(self):
        self.browser.quit()
        if self.auto_save:
            self.save()

    def set_save_path(self, save_path: str):
        """Sets the path to save the file

        Args:
            save_path (str): The saving path
        """
        self.save_path = save_path or datetime.now().strftime("%Y_%m_%d_%H_%M_%S.csv")
        self.file_type = save_path.split(".")[-1] if save_path else "csv"

    def save(self) -> bool:
        """Saves the conversation."""
        save_func = save_func_map.get(self.file_type, None)
        if save_func:
            save_func = getattr(self.chat_history, save_func)
            save_func(self.save_path)
            logging.info("File saved to %s", self.save_path)
            return True

        logging.error("Unsupported file type %s", self.file_type)
        return False

    def find_or_fail(
        self,
        by: By,
        elem_query: str,
        return_type: str = "first",
        return_shadow: bool = False,
        fail_ok: bool = False,
        dom_element: WebElement = None,
    ):
        """Finds a list of elements given elem_query, if none of the items exists, throws an error

        Args:
            by (selenium.webdriver.common.by.By): The method used to locate the element.
            elem_query (str): The elem_query string to locate the element.
            return_type (str): first|all|last. Return first element, all elements or the last one.
            fail_ok (bool): Do not produce error if it is ok to fail.
            dom_element (WebElement): If set, finds within that element.
        Returns:
            WebElement: Web element or None if not found.
        """

        if return_type not in {"first", "last", "all"}:
            return ValueError("Unrecognized return type")

        if dom_element is None:
            dom_element = self.browser.find_elements(by, elem_query)
        else:
            dom_element = dom_element.find_elements(by, elem_query)

        if not dom_element:
            if not fail_ok:
                logging.error(
                    "%s is not located. Please raise an issue with verbose=True", elem_query
                )
            else:
                logging.info("%s is not located.", elem_query)
            return None

        logging.info("%s is located.", elem_query)

        element = {
            "first": lambda x: x[0],
            "all": lambda x: x,
            "last": lambda x: x[-1],
        }[return_type](dom_element)

        if return_shadow:
            return element.shadow_root
        return element

    def check_login_page(self):
        """
        Checks if the login page is displayed in the browser.

        Returns:
            bool: True if the login page is not present, False otherwise.
        """
        login_button = self.browser.find_elements(By.XPATH, self.markers.login_xq)
        return len(login_button) == 0

    def wait_until_appear(self, by: By, elem_query: str) -> Union[WebElement, None]:
        """
        Waits until the specified web element appears on the page.

        This function continuously checks for the presence of a web element.
        It waits until the element is present on the page.
        Once the element has appeared, the function returns.

        Args:
            by (selenium.webdriver.common.by.By): The method used to locate the element.
            elem_query (str): The elem_query string to locate the element.
            timeout_dur (int, optional): Waiting time before the timeout. Default: 15.

        Returns:
            (WebElement | None): If element is appeared, it returns the element. Otherwise,
            it returns None.
        """
        logging.info("Waiting element %s to appear.", elem_query)
        element = None
        try:
            element = self.wait_object.until(
                EC.presence_of_element_located((by, elem_query))
            )
            logging.info("Element %s appeared.", elem_query)
        except Exceptions.TimeoutException:
            logging.info("Element %s is not present, something is wrong.", elem_query)
        return element

    def wait_until_disappear(self, by: By, elem_query: str) -> bool:
        """
        Waits until the specified web element disappears from the page.

        This function continuously checks for the presence of a web element.
        It waits until the element is no longer present on the page.
        Once the element has disappeared, the function returns.

        Args:
            by (selenium.webdriver.common.by.By): The method used to locate the element.
            elem_query (str): The elem_query string to locate the element.
            timeout_dur (int, optional): Waiting time before the timeout. Default: 15.

        Returns:
            (bool) : True if element disappears, false otherwise.
        """
        logging.info("Waiting element %s to disappear.", elem_query)
        try:
            self.wait_object.until_not(EC.presence_of_element_located((by, elem_query)))
            logging.info("Element %s disappeared.", elem_query)
            return True
        except Exceptions.TimeoutException:
            logging.info("Element %s still here, something is wrong.", elem_query)
            return False

    def log_chat(
        self, prompt: str = None, answer: str = None, regenerated: bool = False
    ) -> bool:
        """
        Log a chat interaction in the chat history.

        Parameters:
        - prompt (str): The user's prompt to be logged.
        - answer (str): The answer to the user's prompt to be logged.

        Returns:
        - bool: True if the interaction is logged, False otherwise.
        """
        if not self.auto_save:
            return False

        if prompt:
            self.chat_history.loc[len(self.chat_history)] = [
                "user",
                regenerated,
                prompt,
            ]
        if answer:
            self.chat_history.loc[len(self.chat_history)] = [
                self.client_name,
                regenerated,
                answer,
            ]
        return True

    def preload_custom_func(self) -> None:
        """
        A function to implement specific instructions before loading the webpage
        """
        logging.info(
            """
            The preload behavior is not implemented,
            which could be considered normal if verification is not required.
            """
        )

    def postload_custom_func(self) -> None:
        """
        A function to implement specific instructions after loading the webpage
        """
        logging.info(
            """
            The postload behavior is not implemented,
            which could be considered normal if verification is not required.
            """
        )

    def pass_verification(self) -> bool:
        """
        Performs the verification process on the page if challenge is present.
        Returns:
            None
        """
        logging.info(
            """
            Pass verification function is not implemented,
            which could be considered normal if verification is not required.
            """
        )
        return True

    @abc.abstractmethod
    def login(self, username: str, password: str) -> bool:
        """
        Performs the login process with the provided username and password.
        """
        logging.warning(
            "If you are creating a custom automation, please implement this method!"
        )

    @abc.abstractmethod
    def interact(self, prompt: str) -> str:
        """
        Abstract function to interact with the language model.
        """
        logging.warning(
            "If you are creating a custom automation, please implement this method!"
        )

    @abc.abstractmethod
    def reset_thread(self) -> bool:
        """
        Abstract function to open a new thread.
        """
        logging.warning("Resetting thread is either not implemented or not available")

    @abc.abstractmethod
    def regenerate_response(self) -> str:
        """
        Abstract function to regenerate the answers.
        """
        logging.warning(
            "Regenerating response is either not implemented or not available"
        )

    @abc.abstractmethod
    def switch_model(self, model_name: str) -> bool:
        """
        Abstract function to switch the model.

        Args:
            model_name: (str) = The name of the model.
        """
        logging.warning("Switching model is either not implemented or not available")



"""Storage of the xpath, class and id identifiers"""
from easydict import EasyDict

markers = EasyDict(
    {
        "ChatGPT": {
            "login_xq": '//button[//div[text()="Log in"]]',
            "continue_xq": '//button[text()="Continue"]',
            "tutorial_xq": '//div[contains(text(), "Okay, let’s go")]',
            "button_tq": "button",
            "done_xq": '//button[//div[text()="Done"]]',
            "menu_xq": '//button[contains(@id, "headlessui-menu-button")]',
            "custom_xq": '//a[contains(text(), "Custom")]',
            "custom_toggle_xq": '//button[@role="switch"]',
            "custom_textarea_xq": '//textarea[@type="button"]',
            "custom_save_xq": '//div[contains(text(), "Save")]',
            "custom_cancel_xq": '//div[contains(text(), "Cancel")]',
            "custom_tutorial_xq": '//div[text()="OK"]',
            "chatbox_xq": '//div[@data-message-author-role="assistant"]',
            "wait_xq": '//button[@aria-label="Stop generating"]',
            "reset_xq": '//a[//span[text()="New chat"]]',
            "reset_cq": "truncate",
            "regen_xq": '//button[contains(@class, "group-[.final-completion]:visible")]',
            "textarea_tq": "textarea",
            "textarea_iq": "prompt-textarea",
            "gpt_xq": '//span[text()="{}"]',
        },
        "HuggingChat": {
            "login_xq": '//form[@action="/chat/login"]',
            "username_xq": '//input[@name="username"]',
            "password_xq": '//input[@name="password"]',
            "a_login_xq": '//button[contains(text(), "Login")]',
            "textarea_xq": '//textarea[@enterkeyhint="send"]',
            "stop_gen_xq": '//button[contains(text(),"Stop generating")]',
            "chatbox_xq": '//div[@role="presentation"]',
            "search_xq": '//div[@aria-label="web search toggle"]',
            "model_xq": '//div[div/div/text()="Current Model"]//a',
            "settings_xq": '//h2[text()="Settings"]',
            "model_li_xq": '//div/a/div[@class="truncate"]',
            "model_a_xq": '//div[h2/text()="Settings"]//button',
        },
        "Gemini": {
            "textarea_xq": '//div[@role="textbox"]',
            "wait_xq": '//img[contains(@src, "sparkle_processing")]',
            "chatbox_tq": "message-content",
            "new_chat_xq": '//span[text()="New chat"]',
            "regen_1_xq": '//span[text()="View other drafts"]',
            "regen_2_xq": '//button[@mattooltip="Regenerate drafts"]',
        },
        "Pi": {
            "textarea_xq": '//textarea[@role="textbox"]',
            "sendkeys_xq": '//button[@aria-label="Submit text"]',
            "wait_xq": "//button[@disabled]",
            "chatbox_cq": "break-anywhere",
            "model_1_xq": '//div[contains(@class, "shadow-input")]//button',
            "model_2_xq": '//button[contains(@class, "t-body-s")]',
            "model_v_xq": '//div[contains(text(), "Switched to")]',
        },
        "Copilot": {
            "main_area_tq": "cib-serp",
            "action_bar_iq": "cib-action-bar-main",
            "input_bar_cq": "input-row",
            "textarea_iq": "searchbox",
            "button_cq": "submit",
            "con_main_iq": "cib-conversation-main",
            "con_chat_sq": "cib-chat-turn",
            "con_resp_cq": "response-message-group",
            "con_msg_sq": "cib-message",
            "con_ins_sq": "cib-shared",
            "con_last_cq": "ac-textBlock",
            "ref_cq": "tooltip-target",
            "welcome_tq": "cib-welcome-container",
            "tone_tq": "cib-tone-selector",
            "side_tq": "cib-side-panel",
            "plugin_tq": "cib-plugin-panel",
            "p_control_cq": "plugin-control",
            "new_chat_cq": "button-compose-content",
        },
        "Claude": {
            "start_button_xq": '//div[text()="Start a new chat"]',
            "textarea_cq": "ProseMirror",
            "send_button_xq": '//button[@aria-label="Send Message"]',
            "chatarea_xq": '//div[contains(@class, "grid-cols-1")]/div[@class="contents"]',
            "regen_xq": "//button[text()='Retry']",
        },
    }
)



from .base_multiagent import BaseMultiAgent

class Conversation(BaseMultiAgent):
    
    def __init__(self):
        pass

    def start_conversation(self, text_1: str, text_2: str, use_response_1: bool= True):
        """Starts a conversation between two heads"""

        f_response = self.interact(0, text_1)
        text_2 = text_2 + f_response if use_response_1 else text_2
        s_response = self.interact(1, text_2)

        self.head_responses[0].append(f_response)
        self.head_responses[1].append(s_response)

        return f_response, s_response

    def continue_conversation(self, text_1: str= None, text_2: str= None):
        """Make another round of conversation.
        If text_1 or text_2 is given, the response is not used"""
        text_1 = text_1 or self.head_responses[1][-1]

        f_response = self.interact(0, text_1)
        text_2 = text_2 or f_response

        s_response = self.interact(1, text_2)

        self.head_responses[0].append(f_response)
        self.head_responses[1].append(s_response)
        return f_response, s_response


"""Class definition for GeminiClient"""
import logging
from typing import Union

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from talkingheads.base_browser import BaseBrowser


class GeminiClient(BaseBrowser):
    """GeminiClient class to interact with Gemini"""

    def __init__(self, **kwargs):
        super().__init__(
            client_name="Gemini",
            url="https://gemini.google.com",
            credential_check=False,
            **kwargs,
        )

    def login(self, username: str, password: str) -> bool:
        """
        Performs the login process with the provided username and password.

        This function operates on the login page.
        It finds and clicks the login button,
        fills in the email and password textboxes

        Args:
            username (str): The username to be entered.
            password (str): The password to be entered.

        Returns:
            bool : True if login succesful, False otherwise.
        """
        logging.info(
            "It is not possible to provide login functionality for Google"
            "Please follow the instructions on the repo to connect Gemini"
        )
        text_area = self.find_or_fail(By.XPATH, self.markers.textarea_xq)
        if not text_area and not self.headless:
            for _ in range(5):
                logging.error(
                    """Prompt area can\'t located, use browser to manually
                    login your account, navigate to https://Gemini.google.com/chat
                    and press any key here."""
                )
                input()
                text_area = self.find_or_fail(By.XPATH, self.markers.textarea_xq)
                break
            else:
                logging.error("Login is unsuccesful")
                return False
        return True

    def interact(self, prompt: str) -> str:
        """
        Sends a prompt and retrieves the answer from the ChatGPT system.

        This function interacts with the Gemini.
        It takes the prompt as input and sends it to the system.
        The prompt may contain multiple lines separated by '\\n'.
        In this case, the function simulates pressing SHIFT+ENTER for each line.
        Upon arrival of the interaction, the function waits for the answer.
        Once the response is ready, the function will return the response.

        Args:
            prompt (str): The interaction text.

        Returns:
            str: The generated answer.
        """

        text_area = self.find_or_fail(By.XPATH, self.markers.textarea_xq)
        if not text_area:
            return ""

        for each_line in prompt.split("\n"):
            text_area.send_keys(each_line)
            text_area.send_keys(Keys.SHIFT + Keys.ENTER)
        text_area.send_keys(Keys.RETURN)
        logging.info("Message sent, waiting for response")
        self.wait_until_disappear(By.XPATH, self.markers.wait_xq)
        answer = self.find_or_fail(
            By.TAG_NAME, self.markers.chatbox_tq, return_type="last"
        )
        if not answer:
            logging.info("Answer is not found.")
            return ""

        logging.info("Answer is ready")
        self.log_chat(prompt=prompt, answer=answer.text)
        return answer.text

    def reset_thread(self) -> bool:
        """Function to close the current thread and start new one

        Returns:
            bool: True new chat button is clicked, false otherwise
        """
        new_chat_button = self.find_or_fail(By.XPATH, self.markers.new_chat_xq)
        if not new_chat_button:
            return False

        new_chat_button.click()
        logging.info("New chat is ready")
        return True

    def toggle_search_web(self) -> Union[bool, None]:
        """Function to enable/disable web search feature

        Returns:
            [bool, None] : The status of the web search functionality, None if toggle is not found.
        """
        search_web_toggle = self.find_or_fail(By.XPATH, self.markers.search_xq)
        if not search_web_toggle:
            return None
        search_web_toggle.click()
        state = search_web_toggle.get_attribute("aria-checked")
        state = state == "true"
        logging.info("Search web is %s", ["disabled", "enabled"][state])
        return state

    def regenerate_response(self) -> str:
        """Closes the current thread and starts a new one.

        Args:
            None

        Returns:
            str: The regenerated answer or empty string in case of failure.
        """
        view_drafts = self.find_or_fail(By.XPATH, self.markers.regen_1_xq)
        if not view_drafts:
            return
        view_drafts.click()
        logging.info("Clicked View drafts button")

        regen_button = self.find_or_fail(By.XPATH, self.markers.regen_2_xq)
        if not regen_button:
            return ""

        regen_button.click()
        logging.info("Clicked regenerate button")
        self.wait_until_disappear(By.XPATH, self.markers.wait_xq)
        answer = self.browser.find_elements(By.TAG_NAME, self.markers.chatbox_tq)[-1]
        logging.info("New answer is ready")

        if self.auto_save:
            self.chat_history.loc[len(self.chat_history)] = [
                self.client_name,
                True,
                answer.text,
            ]
        return answer.text

    def switch_model(self, model_name: str) -> bool:
        logging.info("Gemini doesn't have a model selection")
        return False



"""Class definition for PI client"""
import logging
import re
import time

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from ..base_browser import BaseBrowser


class PiClient(BaseBrowser):
    """
    PiClient class to interact with Pi.
    It helps you to conncet to https://pi.ai/.
    Apart from core functionality Pi supports web search.
    It is not possible to regenerate a response by using Pi
    """

    def __init__(self, **kwargs):
        super().__init__(
            client_name="Pi", url="https://pi.ai/talk", credential_check=False, **kwargs
        )

    def login(self, username: str = None, password: str = None):
        """
        Performs the login process with the provided username and password.
        You don't need to login to use Pi

        This function operates on the login page.
        It finds and clicks the login button,
        fills in the email and password textboxes

        Args:
            username (str): The username to be entered.
            password (str): The password to be entered.

        Returns:
            bool : True
        """
        logging.info("Login is not provided for Pi at the moment.")
        return True

    def postload_custom_func(self) -> None:
        """Pi starts with a welcome message, we should wait until the message to finish."""
        time.sleep(2)
        self.browser.get(self.url)
        self.is_ready_to_prompt()
        return

    def is_ready_to_prompt(self) -> bool:
        """
        Checks if the Pi is ready to be prompted.
        The indication for an ongoing message generation process
        is a disabled send button. The indication for no input is the same
        disabled button. Therefore we put a dummy dot into the textarea
        and we are left with the only reason for the button to be disabled,
        that is, a message being generated.

        Returns:
            bool : return if the system is ready to be prompted.
        """

        text_area = self.wait_until_appear(By.XPATH, self.markers.textarea_xq)
        if not text_area:
            return False
        text_area.send_keys(".")
        self.wait_until_disappear(By.XPATH, self.markers.wait_xq)

        # Then, we clear the text area to make space for new interacton :)
        text_area.send_keys(Keys.CONTROL + "a")
        time.sleep(0.1)
        text_area.send_keys(Keys.DELETE)
        time.sleep(0.1)
        return True

    def interact(self, prompt: str):
        """Sends a prompt and retrieves the answer from the ChatGPT system.

        This function interacts with the PI.
        It takes the prompt as input and sends it to the system.
        The prompt may contain multiple lines separated by '\\n'.
        In this case, the function simulates pressing SHIFT+ENTER for each line.
        Upon arrival of the interaction, the function waits for the answer.
        Once the response is ready, the function will return the response.

        Args:
            prompt (str): The interaction text.

        Returns:
            str: The generated answer.
        """

        text_area = self.find_or_fail(By.XPATH, self.markers.textarea_xq)
        if not text_area:
            return ""

        for each_line in prompt.split("\n"):
            text_area.send_keys(each_line)
            text_area.send_keys(Keys.SHIFT + Keys.ENTER)

        self.find_or_fail(By.XPATH, self.markers.sendkeys_xq).click()
        logging.info("Message sent, waiting for response")

        if not self.is_ready_to_prompt():
            return False

        answer = self.find_or_fail(
            By.CLASS_NAME, self.markers.chatbox_cq, return_type="last"
        )
        if not answer:
            return ""
        logging.info("Answer is ready")
        return answer.text

    def reset_thread(self) -> bool:
        """
        Function to close the current thread and start new one

        Returns:
            bool: False always, it is not possible to reset in Pi.
        """
        logging.info("Pi doesn't provide a way to reset the thread")
        return False

    def switch_model(self, model_name: str):
        """
        Switch the model.

        Args:
            model_name: str = The name of the model

        Returns:
            bool: True on success, False on fail
        """
        model_button = self.find_or_fail(By.XPATH, self.markers.model_1_xq)
        if not model_button:
            return False
        model_button.click()
        time.sleep(1)

        models = self.find_or_fail(By.XPATH, self.markers.model_2_xq, return_type="all")
        if not models:
            return False
        models = {model.text or "Pi": model for model in models}
        logging.info(models.keys())

        model = models.get(model_name, None)
        if model is None:
            logging.error("Model %s has not found", model_name)
            logging.error("Available models are: %s", str(models.keys()))
            return False
        model.click()

        verification = self.find_or_fail(By.XPATH, self.markers.model_v_xq)
        if not re.search(rf"Switched to( just)? {model_name}", verification.text):
            logging.error("Model switch to %s is unsuccessful", model_name)
            return False

        logging.info("Switched to %s", model_name)
        return True

    def regenerate_response(self):
        raise NotImplementedError("Pi doesn't provide response regeneration")
